C*******************************************************************
C     COMPUTES GRADIENT BY HIGH PRECISION NUMERICAL DIFFERENTIATION
C*******************************************************************
C COPYRIGHT = SPELLUCCI
      SUBROUTINE EGRADF(XL,GRADXL)
      INCLUDE 'O8FUCO.INC'
      INCLUDE 'O8CONS.INC'
      DOUBLE PRECISION XL(*),GRADXL(*),FXL(6)
      DOUBLE PRECISION SD1,SD2,SD3,XINCR,ZZ,D1,D2,D3
      INTEGER I
C****************************************************
C     HIGH PRECISION NUMERICAL DIFFERENTIATION
C     BY SIXTH ORDER EXTRAPOLATION
C     USES RICHARDSON-EXTRAPOLATION OF THREE VALUES
C     OF THE SYMMETRIC DIFFERENCE QUOTIENT WITH
C     RATHER LARGE DISCRETIZATION STEPSIZE
C     THE OPTIMAL STEPSIZE DEPENDS ON THE SEVENTH
C     PARTIAL DERIVATIVES OF THE FUNCTION . SINCE THESE
C     ARE INACCESSIBLE, WE REPLACE THEM BY ONE IN THE
C     ESTIMATE OF DELDIF
C****************************************************
      DELDIF=EXP(LOG(EPSMAC)/SEVEN)*TWOM2
      ICGF=ICGF+1
      DO   I=1,N
        ZZ=XL(I)
        XINCR=DELDIF*(ABS(XL(I))+ONE)
        XL(I)=ZZ-XINCR
        CALL EF(XL,FXL(1))
        XL(I)=ZZ+XINCR
        CALL EF(XL,FXL(2))
        XINCR=XINCR+XINCR
        D1=XINCR
        XL(I)=ZZ-XINCR
        CALL EF(XL,FXL(3))
        XL(I)=ZZ+XINCR
        CALL EF(XL,FXL(4))
        XINCR=XINCR+XINCR
        D2=XINCR
        XL(I)=ZZ-XINCR
        CALL EF(XL,FXL(5))
        XL(I)=ZZ+XINCR
        CALL EF(XL,FXL(6))
        XL(I)=ZZ
        D3=XINCR+XINCR
        SD1=(FXL(2)-FXL(1))/D1
        SD2=(FXL(4)-FXL(3))/D2
        SD3=(FXL(6)-FXL(5))/D3
        SD3=SD2-SD3
        SD2=SD1-SD2
        SD3=SD2-SD3
        GRADXL(I)=SD1+P4*SD2+SD3/C45
      ENDDO
      RETURN
      END
C*******************************************************************
      SUBROUTINE EGRADG(J,XL,GRADXL)
      INCLUDE 'O8FUCO.INC'
      INCLUDE 'O8CONS.INC'
      DOUBLE PRECISION XL(*),GRADXL(*),FXL(6)
      DOUBLE PRECISION SD1,SD2,SD3,XINCR,ZZ,D1,D2,D3
      INTEGER I,J
C****************************************************
C     HIGH PRECISION NUMERICAL DIFFERENTIATION
C     BY SIXTH ORDER EXTRAPOLATION
C     USES RICHARDSON-EXTRAPOLATION OF THREE VALUES
C     OF THE SYMMETRIC DIFFERENCE QUOTIENT WITH
C     RATHER LARGE DISCRETIZATION STEPSIZE
C     THE OPTIMAL STEPSIZE DEPENDS ON THE SEVENTH
C     PARTIAL DERIVATIVES OF THE FUNCTION . SINCE THESE
C     ARE INACCESSIBLE, WE REPLACE THEM BY ONE IN THE
C     ESTIMATE OF DELDIF
C****************************************************
      DELDIF=EXP(LOG(EPSMAC)/SEVEN)*TWOM2
      CGRES(J+NH)=CGRES(J+NH)+1
      DO   I=1,N
        ZZ=XL(I)
        XINCR=DELDIF*(ABS(XL(I))+ONE)
        XL(I)=ZZ-XINCR
        CALL EG(J,XL,FXL(1))
        XL(I)=ZZ+XINCR
        CALL EG(J,XL,FXL(2))
        XINCR=XINCR+XINCR
        D1=XINCR
        XL(I)=ZZ-XINCR
        CALL EG(J,XL,FXL(3))
        XL(I)=ZZ+XINCR
        CALL EG(J,XL,FXL(4))
        XINCR=XINCR+XINCR
        D2=XINCR
        XL(I)=ZZ-XINCR
        CALL EG(J,XL,FXL(5))
        XL(I)=ZZ+XINCR
        CALL EG(J,XL,FXL(6))
        XL(I)=ZZ
        D3=XINCR+XINCR
        SD1=(FXL(2)-FXL(1))/D1
        SD2=(FXL(4)-FXL(3))/D2
        SD3=(FXL(6)-FXL(5))/D3
        SD3=SD2-SD3
        SD2=SD1-SD2
        SD3=SD2-SD3
        GRADXL(I)=SD1+P4*SD2+SD3/C45
      ENDDO
      RETURN
      END
C*******************************************************************
      SUBROUTINE EGRADH(J,XL,GRADXL)
      INCLUDE 'O8FUCO.INC'
      INCLUDE 'O8CONS.INC'
      DOUBLE PRECISION XL(*),GRADXL(*),FXL(6)
      DOUBLE PRECISION SD1,SD2,SD3,XINCR,ZZ,D1,D2,D3
      INTEGER I,J
C****************************************************
C     HIGH PRECISION NUMERICAL DIFFERENTIATION
C     BY SIXTH ORDER EXTRAPOLATION
C     USES RICHARDSON-EXTRAPOLATION OF THREE VALUES
C     OF THE SYMMETRIC DIFFERENCE QUOTIENT WITH
C     RATHER LARGE DISCRETIZATION STEPSIZE
C     THE OPTIMAL STEPSIZE DEPENDS ON THE SEVENTH
C     PARTIAL DERIVATIVES OF THE FUNCTION . SINCE THESE
C     ARE INACCESSIBLE, WE REPLACE THEM BY ONE IN THE
C     ESTIMATE OF DELDIF
C****************************************************
      DELDIF=EXP(LOG(EPSMAC)/SEVEN)*TWOM2
      CGRES(J)=CGRES(J)+1
      DO   I=1,N
        ZZ=XL(I)
        XINCR=DELDIF*(ABS(XL(I))+ONE)
        XL(I)=ZZ-XINCR
        CALL EH(J,XL,FXL(1))
        XL(I)=ZZ+XINCR
        CALL EH(J,XL,FXL(2))
        XINCR=XINCR+XINCR
        D1=XINCR
        XL(I)=ZZ-XINCR
        CALL EH(J,XL,FXL(3))
        XL(I)=ZZ+XINCR
        CALL EH(J,XL,FXL(4))
        XINCR=XINCR+XINCR
        D2=XINCR
        XL(I)=ZZ-XINCR
        CALL EH(J,XL,FXL(5))
        XL(I)=ZZ+XINCR
        CALL EH(J,XL,FXL(6))
        XL(I)=ZZ
        D3=XINCR+XINCR
        SD1=(FXL(2)-FXL(1))/D1
        SD2=(FXL(4)-FXL(3))/D2
        SD3=(FXL(6)-FXL(5))/D3
        SD3=SD2-SD3
        SD2=SD1-SD2
        SD3=SD2-SD3
        GRADXL(I)=SD1+P4*SD2+SD3/C45
      ENDDO
      RETURN
      END

